#include "IZ_app.h"

u64 IZ_AppGetTicks(struct IZ_App* app) {
	return app->ticks;
}

void IZ_AppBindConnection(struct IZ_App* app, struct lws* wsi) {
	app->net_state.ws.connection = wsi;
}

IZ_NetState* IZ_AppGetNetState(struct IZ_App* app) {
	return &app->net_state;
}

IZ_InputState* IZ_AppGetInputState(struct IZ_App* app) {
	return &app->input_state;
}

IZ_ProcedureResult IZ_AppInitialize(struct IZ_App* app, u8 argc, const char* argv[]) {
	memset(app, 0, sizeof(struct IZ_App));
	u32 flags = (
		SDL_INIT_VIDEO
		| SDL_INIT_GAMECONTROLLER
		| SDL_INIT_EVENTS
	);

	if (SDL_Init(flags) < 0) {
		SDL_LogError(SDL_LOG_CATEGORY_ERROR, "SDL could not initialize! SDL_Error: %s\n", SDL_GetError());
		return IZ_APP_RUN_SDL_INIT_ERROR;
	}

	const char* cmdline_buffer;
	char config_path[128];
	// TODO abstract command line args parsing
	if ((cmdline_buffer = IZ_ConfigGetCommandlineOption(argc, argv, "-c"))) {
		memcpy_s(config_path, 128, cmdline_buffer, 128);
	} else {
		IZ_ConfigGetDefaultPath(config_path, 128);
	}

	if (IZ_VideoInitialize(&app->video_state, app, config_path, argc, argv)) {
		return IZ_APP_RUN_VIDEO_INIT_ERROR;
	}

	if (IZ_InputInitialize(&app->input_state, config_path, argc, argv)) {
		return IZ_APP_RUN_INPUT_INIT_ERROR;
	}

	if (IZ_NetInitialize(&app->net_state, app, IZ_AppConnect, config_path, argc, argv)) {
		return IZ_APP_RUN_NETWORKING_ERROR;
	}

	IZ_PoolInitialize(&app->pool, POOL_MAX_SIZE);

  // TODO put into its timer module
	app->ticks = 0;
	return IZ_APP_RUN_RESULT_OK;
}

void IZ_AppTeardown(struct IZ_App* app) {
	IZ_NetDisconnect(&app->net_state);
	IZ_PoolTeardown(&app->pool);
	IZ_InputTeardown(&app->input_state);
	IZ_VideoTeardown(&app->video_state);
	SDL_Quit();
}

IZ_ProcedureResult IZ_AppHandleSDLEvents(struct IZ_App* app) {
	SDL_Event e;
	while (SDL_PollEvent(&e) != 0) {
		if (e.type == SDL_QUIT) {
			return 1;
		}

		if (e.type == SDL_KEYDOWN) {
			if (e.key.keysym.sym == SDLK_PAGEUP) {
				IZ_NetConnect(
					&app->net_state,
					(IZ_WSClientInitializeParams) {
						.host = "127.0.0.1",
						.path = "/",
						.port = 42069,
					}
				);
			} else if (e.key.keysym.sym == SDLK_PAGEDOWN) {
				IZ_NetDisconnect(&app->net_state);
			} else if (e.key.keysym.sym == SDLK_INSERT) {
				IZ_NetSendTextMessage(&app->net_state, "hello", 5);
			}
		}

		IZ_InputHandleSDLEvents(&app->input_state, e);
	}
	return 0;
}

void IZ_AppHandlePortMIDIEvents(struct IZ_App* app) {
	u8 player_index;
	i32* midi_events_count;
	u32 midi_event_index;
	for (player_index = 0; player_index < IZ_PLAYERS; player_index += 1) {
		if (!app->input_state.midi_input_state[player_index].device_info) {
			continue;
		}

		midi_events_count = &app->input_state.midi_input_state[player_index].midi_events_count;
		*midi_events_count = Pm_Read(
			app->input_state.midi_input_state[player_index].stream,
			// TODO bind buffers and streams to device instead of player input state
			app->input_state.midi_input_state[player_index].event_buffer,
			1024
		);

		if (*midi_events_count < 1) {
			continue;
		}

		for (midi_event_index = 0; midi_event_index < *midi_events_count; midi_event_index += 1) {
			IZ_InputHandlePortMIDIEvents(
				&app->input_state,
				app->input_state.midi_input_state[player_index].event_buffer[midi_event_index]
			);
		}
	}
}

IZ_ProcedureResult IZ_AppHandleInputEvents(struct IZ_App* app) {
	i32 sdl_events_result = IZ_AppHandleSDLEvents(app);
	if (sdl_events_result) {
		return sdl_events_result;
	}

	IZ_AppHandlePortMIDIEvents(app);
	return 0;
}

void IZ_VideoUpdateForDebugTicks(IZ_VideoState* video_state, uint64_t ticks) {
	SDL_SetRenderDrawColor(video_state->renderer, 0x00, 0xff, 0xff, 0xff);
	u64 the_ticks = ticks;
	u8 column;
	u8 row;
	const u8 size = 4;

	u8 i;
	for (i = 0; i < 64; i += 1) {
		column = i % 32;
		row = i / 32;

		if (the_ticks & 0x1) {
			SDL_RenderFillRectF(video_state->renderer, &(SDL_FRect) {
				(f32) (video_state->config.width - ((column + 1) * size)),
				(f32) (video_state->config.height - ((row + 1) * size)),
				size,
				size
			});
		}
		the_ticks >>= 1;
	}
}

void IZ_VideoUpdateForDebugInput(IZ_VideoState* video_state, IZ_InputState* input_state) {
	SDL_SetRenderDrawColor(video_state->renderer, 0xff, 0xff, 0x00, 0xff);
	const u8 size = 4;

	u8 column;
	u8 row;

	u8 p;
	u8 i;
	for (p = 0; p < IZ_PLAYERS; p += 1) {
		IZ_Action the_action = input_state->action[p];
		for (i = 0; i < CONTROLS; i += 1) {
			column = (i % 4) + (p * 4);
			row = i / 4;

			if (the_action & 0x1) {
				SDL_RenderFillRectF(video_state->renderer, &(SDL_FRect) {
					(f32) (column * size),
					(f32) (row * size),
					size,
					size
				});
			}
			the_action >>= 1;
		}
	}
}

void IZ_VideoUpdateForDebugNet(IZ_VideoState* video_state, IZ_NetState* net_state) {
	if (!net_state->ws.connection) {
		return;
	}
	const u8 size = 4;

	SDL_SetRenderDrawColor(video_state->renderer, 0x00, 0xff, 0x00, 0xff);
	SDL_RenderFillRectF(video_state->renderer, &(SDL_FRect) {
		0,
		(f32) (video_state->config.height - size),
		size,
		size,
	});

	u8 column;
	u8 row;

	u8 p;
	u8 i;
	for (p = 0; p < IZ_PLAYERS; p += 1) {
		IZ_Action the_action = net_state->action[p];
		for (i = 0; i < CONTROLS; i += 1) {
			column = (i % 4) + (p * 4);
			row = i / 4;

			if (the_action & 0x1) {
				SDL_RenderFillRectF(video_state->renderer, &(SDL_FRect) {
					(f32) (column * size),
					(f32) ((row * size) + (video_state->config.height - (size * 5))),
					size,
					size
				});
			}
			the_action >>= 1;
		}
	}
}

void IZ_VideoUpdateForDebug(IZ_VideoState* video_state, u64 ticks, IZ_InputState* input_state, IZ_NetState* net_state) {
	IZ_VideoUpdateForDebugTicks(video_state, ticks);
	IZ_VideoUpdateForDebugInput(video_state, input_state);
	IZ_VideoUpdateForDebugNet(video_state, net_state);
}

void IZ_VideoUpdate(IZ_VideoState* video_state) {
	struct IZ_App* app = video_state->user_data;
	if (app->ticks - video_state->last_update_at > 1000 / video_state->config.max_fps) {
		// Update window
		SDL_SetRenderDrawColor(video_state->renderer, 0x00, 0x00, 0x00, 0xff);
		SDL_RenderClear(video_state->renderer);
		for (u8 i = 0; i < MAX_ACTIVE_SPRITES; i += 1) {
			if (!video_state->active_sprites[i]) {
				continue;
			}
			// TODO draw sprites
		}
		IZ_VideoUpdateForDebug(video_state, app->ticks, &app->input_state, &app->net_state);
		SDL_RenderPresent(video_state->renderer);
		video_state->last_update_at = app->ticks;
	}
}

IZ_ProcedureResult IZ_AppRun(struct IZ_App* app, u8 argc, const char* argv[]) {
	IZ_ProcedureResult init_result = IZ_AppInitialize(app, argc, argv);
	if (init_result) {
		return init_result;
	}

	while (true) {
		app->ticks = SDL_GetTicks64();

		// TODO do audio processing
		// TODO do networking

		if (IZ_AppHandleInputEvents(app)) {
			break;
		}

		IZ_AppHandleOutboundNetworking(app);
		IZ_VideoUpdate(&app->video_state);
	}

	IZ_AppTeardown(app);
	return IZ_APP_RUN_RESULT_OK;
}
